# pylint: disable=C0114
from __future__ import annotations
from typing import NamedTuple
from collections.abc import Iterator
import itertools

from fontParts.fontshell import RComponent
from fontParts.fontshell import RContour
from smufolib.objects.glyph import Glyph
from smufolib.objects.layer import Layer
from smufolib.objects.font import Font
from smufolib import error, stdUtils

TYPES = ('line', 'curve', 'qcurve')

# pylint: disable=C0103


class Point(NamedTuple):
    """Named tuple for point values.

    :param type: The :attr:`~fontParts.base.BasePoint.type` of the point.
    :param position: The :class:`~Position` of the point.
    :param contourIndex: The :attr:`~fontParts.base.BaseContour.index` of
        the point's parent contour.

    """
    type: str
    position: Position
    contourIndex: int


class Position(NamedTuple):
    """Named tuple for position values.

    :param x: The horizontal position of the point.
    :param y: The vertical position of the point.

    """
    x: int | float
    y: int | float


def getPoints(obj: Glyph | tuple[Glyph, ...] | Layer | Font,
              types: str | tuple[str, ...] = TYPES
              ) -> Iterator[Point]:
    """Get points from font-related object.

    This function simply concatinates the results
    of :func:`getContourPoints` and :func:`getCompositePoints` and
    returns the result as an :obj:`itertools.chain` object.

    :param obj: The object from which to get the points. This can be a
        single :class:`.Glyph`, a :class:`tuple` of glyph objects or any
        parent object (:class:`.Layer` or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.
    :raises ValueError: If `obj` items is not an accepted type.

    """
    error.validateType(obj, (tuple, Glyph, Layer, Font), 'obj')
    if isinstance(obj, tuple):
        error.validateType(obj[0], Glyph, 'obj', items=True)

    return itertools.chain(getContourPoints(obj, types),
                           getCompositePoints(obj, types))


def getContourPoints(obj: RContour | tuple[RContour, ...]
                     | Glyph | Layer | Font,
                     types: str | tuple[str, ...] = TYPES
                     ) -> Iterator[Point]:
    """Get contour points from a font-related object.

    :param obj: The object from which to get the points. This can be a
        single contour object, a :class:`tuple` of contours or any parent
        object (:class:`.Glyph`, :class:`.Layer` or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.

    """
    error.validateType(obj, (tuple, RContour, Glyph, Layer, Font), 'obj')
    if isinstance(obj, tuple):
        for component in obj:
            error.validateType(component, RContour, 'obj', items=True)
    if isinstance(obj, RContour):
        obj = (obj,)

    rawPoints = stdUtils.flatten(obj)

    return (Point(p.type, Position(*p.position), p.contour.index)
            for p in rawPoints if p.type in types or p.type == types)


def getCompositePoints(obj: RComponent | tuple[RComponent, ...]
                       | Glyph | Layer | Font,
                       types: str | tuple[str, ...] = TYPES
                       ) -> Iterator[Point]:
    """Get composite points from font-related object.

    The component's offset position is taken into account.

    :param obj: The object from which to get the points. This can be a
        single component object, a :class:`tuple` of components or any
        parent object (:class:`.Glyph`, :class:`.Layer`
        or :class:`.Font`).
    :param types: The :attr:`fontParts.base.BasePoint.type` to include.
        Defaults to ``('line', 'curve', 'qcurve')``.
    :raises TypeError: If `obj` is not an accepted type.

    """
    components = _getComponents(obj)

    def pointGenerator():
        for component in components:
            baseGlyph = component.font[component.baseGlyph]
            rawPoints = stdUtils.flatten(baseGlyph.contours)
            offset, index = component.offset, component.index
            for point in rawPoints:
                if point.type in types or point.type == types:
                    newPosition = stdUtils.addTuples(
                        point.position, offset)
                    yield Point(point.type, Position(*newPosition), index)

    return pointGenerator()


def _getComponents(obj: RComponent
                   | tuple[RComponent, ...] | Glyph | Layer | Font
                   ) -> tuple[RComponent, ...] | Iterator[RComponent]:
    # Get components from font-related object.
    if isinstance(obj, tuple):
        for component in obj:
            error.validateType(component, RComponent, 'obj', items=True)
        return obj
    if isinstance(obj, RComponent):
        return (obj,)
    if isinstance(obj, Glyph):
        return obj.components
    if isinstance(obj, (Layer, Font)):
        return stdUtils.flatten(g.components for g in obj)
    raise TypeError(
        error.generateTypeError(
            obj, (RComponent, tuple, Glyph, Layer, Font), 'obj'
        )
    )



import unittest

from smufolib.objects.font import Font
from smufolib.pointUtils import (
    Point,
    Position,
    getPoints,
    getContourPoints,
    getCompositePoints
)

# pylint: disable=C0115, C0116, C0103


class PointUtils(unittest.TestCase):

    def setUp(self):
        pass

    def test_getPoints(self):
        pass

    def test_getContourPoints(self):
        pass

    def test_getCompositePoints(self):
        pass


if __name__ == '__main__':
    unittest.main()



# pylint: disable=C0114
from __future__ import annotations
from typing import Any
from collections.abc import Generator, Iterable
import operator

# pylint: disable=C0103


def flatten(iterable: Iterable[Any],
            depth: int | None = None
            ) -> Generator:
    """Flatten irregularly nested iterables of any depth.

    :param iterable: The :term:`iterable` to flatten.
    :param depth: The number of levels to flatten. `depth=None` employs
        maximum flattening. Defaults to :obj:`None`.

    """
    for item in iterable:
        if (not isinstance(item, Iterable)
                or isinstance(item, (str, bytes)) or depth == 0):
            yield item
        elif depth is None:
            yield from flatten(item)
        else:
            yield from flatten(item, depth - 1)


def addTuples(*tuples: tuple[int | float, ...]) -> tuple[int | float, ...]:
    r"""Sumize tuple values.

    :param \*tuples: tuples to add.

    Example::

        >>> addTuples((2, 4), (2, 4))
        (4, 8)

    """
    return tuple(map(sum, zip(*tuples)))


def getSummary(docstring: str | None) -> str | None:
    """Get summary line from docstring.

    :param docstring: The docstring from which to get the summary line.

    """
    if docstring is None:
        return None

    return docstring.split('\n', maxsplit=1)[0]


def isFloat(string: str) -> bool:
    """Check if string represents a :class:`float`.

    param string: The string to check.

    """
    if '.' not in string:
        return False
    try:
        float(string)
        return True
    except ValueError:
        return False


def validateClassAttr(obj, attributes: Iterable[str] | None = None) -> bool:
    """Check for class and attribute exsistence.

    Validates object based on its existence and specified attributes.

    :param obj: The object to validate.
    :param attributes: The attribute names to check.

    """
    if isinstance(attributes, str):
        attributes = (attributes,)
    if attributes is None:
        attributes = ()
    getter = operator.attrgetter(*attributes)
    if all(getter(obj)):
        return True
    return False


def doNothing(*args: Any, **kwargs: Any) -> None:
    # pylint: disable=W0613
    r"""Do nothing at all.

    :param \*args: Positional arguments.
    :param \**kwargs: Keyword arguments.

    """


def verbosePrint(message: str, verbose: bool,
                 *args: Any, **kwargs: Any) -> None:
    r"""Print a message if `verbose` is :obj:`True`.

    Behaves like the built-in :func:`print` function, but only prints
    the message if `verbose` is :obj:`True`. If `verbose`
    is :obj:`False`, it does nothing.

    :param message: The message to be printed.
    :param verbose: A flag indicating whether to print the message.
    :param \*args: Additional positional arguments passed to
        the :func:`print` function.
    :param \**kwargs: Additional keyword arguments passed to
        the :func:`print` function.

    """
    if verbose:
        print(message, *args, **kwargs)
    else:
        doNothing(message, *args, **kwargs)



# pylint: disable=C0114
from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable
from pathlib import Path

from fontParts.base.normalizers import normalizeInternalObjectType
from fontParts.base.normalizers import normalizeBoolean  # pylint: disable=W0611
from fontParts.base.normalizers import normalizeColor  # pylint: disable=W0611
from fontParts.base.normalizers import normalizeVisualRounding  # pylint: disable=W0611
from fontParts.base.normalizers import normalizeGlyph  # pylint: disable=W0611
from smufolib import error

if TYPE_CHECKING:
    from smufolib.objects.font import Font
    from smufolib.request import Request
    from smufolib.objects.smufl import Smufl
    from smufolib.objects.engravingDefaults import EngravingDefaults

# Type aliases
EngravingDefaultsInput = int | float | tuple[str, ...] | list[str]
EngravingDefaultsReturn = int | float | tuple[str, ...]

# pylint: disable=C0103, C0415


# ----
# Font
# ----

def normalizeFont(value: Font) -> Font:
    """Normalize Font object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.font import Font
    return normalizeInternalObjectType(value, Font, "Font")


# -----
# Smufl
# -----

def normalizeClasses(value: tuple[str, ...] | None) -> tuple[str, ...] | None:
    """Normalize smufl classes.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError:
        - If any `value` item does not normalize
          with :func:`normalizeSmuflName`.
        - If any `value` items are duplicates.

    """
    if value is None:
        return None

    objectName = 'Smufl.classes'
    error.validateType(value, (tuple, list), objectName)

    for val in value:
        error.validateType(val, str, objectName, items=True)
        for v in val.split('_'):
            normalizeSmuflName(v)

    duplicates = {v for v in value if value.count(v) > 1}
    if len(duplicates) != 0:
        raise ValueError(
            error.generateErrorMessage(
                'duplicateItems', objectName=objectName
            )
        )

    return tuple(v for v in value if v)


def normalizeDescription(value: str | None) -> str | None:
    """Normalize smufl descriptions.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError: If `value` is an empty string.

    """
    if value is None:
        return None

    error.validateType(value, (str, type(None)), 'Smufl.description')

    if len(value) == 0:
        raise ValueError(
            error.generateErrorMessage(
                'emptyValue', objectName='Smufl.description'
            )
        )

    return value


def normalizeDesignSize(value: int | None) -> int | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError: If `value` is less than ``10``.

    """
    error.validateType(value, (int, type(None)), 'Smufl.designSize')

    if value is None:
        return None

    if value < 10:
        raise ValueError(
            error.generateErrorMessage(
                'valueTooLow', objectName='Smufl.designSize', value=10
            )
        )

    return value


def normalizeSizeRange(value: tuple[int, int] | None
                       ) -> tuple[int, int] | None:
    """Normalize design size.

    :param value: The value to normalize.
    :raises TypeError: If `value` or it's contents is not an accepted
        type.
    :raises ValueError:
        - If `value` does not contain a value pair.
        - If `value` is not an increasing range.
        - If `value` items do not normalize
          with :func:`normalizeDesignSize`, except not be :obj:`None`.

    """
    if value is None:
        return None

    objectName = 'Smufl.sizeRange'

    error.validateType(value, (tuple, list, type(None)), objectName)

    if len(value) != 2:
        raise ValueError(
            error.generateErrorMessage(
                'singleItem', objectName='Smufl.sizeRange'
            )
        )

    start, end = value

    startNormalized = normalizeDesignSize(start)
    endNormalized = normalizeDesignSize(end)

    if start >= end:
        raise ValueError(
            error.generateErrorMessage(
                'nonIncreasingRange', objectName=objectName
            )
        )

    return startNormalized, endNormalized  # type: ignore


def normalizeSmufl(value: Smufl) -> Smufl:
    """Normalize Smufl object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.smufl import Smufl
    return normalizeInternalObjectType(value, Smufl, "Smufl")


def normalizeSmuflName(value: str | None) -> str | None:
    """Normalize smufl names.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not an accepted type.
    :raises ValueError:
        - If `value` is an empty string.
        - If `value` is value contains non-alphanumeric characters.
        - If `value` does not start with a lowercase letter or number.

    """
    if value is None:
        return None

    objectName = 'Smufl.name'

    error.validateType(value, (str, type(None)), objectName)
    if not value:
        raise ValueError(
            error.generateErrorMessage(
                'emptyValue', objectName=objectName
            )
        )

    for val in value:
        if not val.isalnum():
            raise ValueError(
                error.generateErrorMessage(
                    'alphanumericValue', objectName=objectName
                )
            )

    if value[0].isupper():
        raise ValueError(
            error.generateErrorMessage(
                'invalidInitialCharacter', objectName=objectName
            )
        )

    return value


# ------------------
# Engraving defaults
# ------------------

def normalizeEngravingDefaults(value: EngravingDefaults) -> EngravingDefaults:
    """Normalize EngravingDefaults object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.objects.engravingDefaults import EngravingDefaults
    return normalizeInternalObjectType(
        value, EngravingDefaults, "EngravingDefaults"
    )


def normalizeEngravingDefaultsAttr(name: str,
                                   value: EngravingDefaultsInput | None
                                   ) -> EngravingDefaultsReturn | None:
    """Normalize engraving defaults attribute value based on name.

    :param name: The name of the attribute to normalize.
    :param value: The value to normalize.
    :raises TypeError: If any parameter value is not an accepted type.
    :raises AttributeError: If name is not a
        valid :class:`EngravingDefaults` attribute name.
    :raises ValueError: If  ``'name=textFontFamily'`` and `value` is not
        a :class:`str` or empty.

    """
    from smufolib.objects.engravingDefaults import ENGRAVING_DEFAULTS_KEYS

    if not isinstance(name, str):
        raise TypeError(
            error.generateTypeError(
                value=name, validTypes=str, objectName='name'
            )
        )

    className = 'EngravingDefaults'
    if name not in ENGRAVING_DEFAULTS_KEYS:
        raise AttributeError(
            error.generateErrorMessage(
                'attributeError',
                objectName=className,
                attribute=name
            )
        )

    if value is None:
        return () if value == 'textFontFamily' else None

    objectName = f'{className}.{name}'
    if name == 'textFontFamily':
        if not isinstance(name, (tuple, list)):
            raise TypeError(
                error.generateTypeError(
                    value=value,
                    validTypes=(tuple, list),
                    objectName=objectName
                )
            )

        value = _normalizeStringTuple(objectName, value)

    elif not isinstance(name, (int, float)):
        raise TypeError(
            error.generateTypeError(
                value=value,
                validTypes=(int, float),
                objectName=objectName
            )
        )

    return value


def _normalizeStringTuple(objectName: str,
                          value: tuple[str, ...] | list[str]
                          ) -> tuple[str, ...]:
    # Normalize string tuple.
    error.validateType(value, (tuple, list), objectName)

    for val in value:
        error.validateType(val, str, objectName, items=True)
        if not val:
            raise ValueError(
                error.generateErrorMessage(
                    'emptyValueItems', objectName=objectName
                )
            )

    return tuple(value)


# -------
# Request
# -------


def normalizeRequest(value: Request) -> Request:
    """Normalize Request object.

    :param value: The value to normalize.
    :raises TypeError: If `value` is not the accepted type.

    """
    from smufolib.request import Request
    return normalizeInternalObjectType(value, Request, "Request")


def normalizeRequestPath(value: Path | str | None,
                         parameter: str
                         ) -> str | None:
    """Normalize Request path.

    Relative paths are resolved automatically.

    :param value: The value to normalize.
    :param parameter: The name of the parameter being validated.
    :raises TypeError: If `value` is not the accepted type.

    """
    if value is None:
        return None
    error.validateType(value, (str, Path), parameter)
    if isinstance(value, Path) or value.startswith('.'):
        return str(Path(value).resolve())
    return str(value)



"""where SMuFL meets UFO

"""
from smufolib._version import __version__
from smufolib.objects.engravingDefaults import EngravingDefaults
from smufolib.objects.engravingDefaults import ENGRAVING_DEFAULTS_KEYS
from smufolib.objects.font import Font
from smufolib.objects.glyph import Glyph
from smufolib.objects.layer import Layer
from smufolib.objects.range import Range
from smufolib.objects.smufl import Smufl
from smufolib.objects.smufl import ANCHOR_NAMES
from smufolib.objects.smufl import FONT_ATTRIBUTES
from smufolib.objects.smufl import GLYPH_ATTRIBUTES
from smufolib.request import Request



import unittest
from smufolib import Font, Glyph, EngravingDefaults
from smufolib.normalizers import (
    normalizeFont,
    normalizeClasses,
    normalizeDescription,
    normalizeDesignSize,
    normalizeSizeRange,
    normalizeSmufl,
    normalizeSmuflName,
    normalizeEngravingDefaults,
    normalizeEngravingDefaultsAttr,
    normalizeRequest,
    normalizeRequestPath
)

# pylint: disable=C0115, C0116, C0103


class Normalizers(unittest.TestCase):
    faultyNames = ['', 'smufl-name42', 'SmuflName42']
    font = Font()
    for name in ['uniE001', 'uniE002', 'uniE003', 'uniE004']:
        font[name] = Glyph()

    def test_normalizeFont(self):
        glyph = self.font['uniE001']
        with self.assertRaises(TypeError):
            normalizeFont(glyph)

    def test_normalizeClasses(self):
        self.assertEqual(normalizeClasses(['className']), ('className',))
        with self.assertRaises(TypeError):
            normalizeClasses('className')
        for faultyValue in [
            [None],
            ['className', 'className'],
            [''],
            ['smufl-name42'],
                ['SmuflName42']]:
            with self.assertRaises(ValueError):
                normalizeClasses(faultyValue)

    def test_normalizeDescription(self):
        with self.assertRaises(TypeError):
            normalizeDescription(0)
        with self.assertRaises(ValueError):
            normalizeDescription('')

    def test_normalizeDesignSize(self):
        with self.assertRaises(TypeError):
            normalizeDesignSize('24')
        with self.assertRaises(ValueError):
            normalizeDesignSize(9)

    def test_normalizeSizeRange(self):
        self.assertEqual(normalizeSizeRange([10, 36]), (10, 36))
        with self.assertRaises(TypeError):
            normalizeSizeRange(10)
        for faultyValue in [[24], [36, 10], [9, 36]]:
            with self.assertRaises(ValueError):
                normalizeSizeRange(faultyValue)

    def test_normalizeSmufl(self):
        with self.assertRaises(TypeError):
            normalizeSmufl(self.font)

    def test_normalizeSmuflName(self):
        with self.assertRaises(TypeError):
            normalizeSmuflName(0)
        for name in self.faultyNames:
            with self.assertRaises(ValueError):
                normalizeSmuflName(name)

    def test_normalizeEngravingDefaults(self):
        with self.assertRaises(TypeError):
            normalizeEngravingDefaults(self.font)

    def test_normalizeEngravingDefaultsAttr(self):
        for wrongAttr, wrongVal in [('attribute', '23'),
                                    ('attribute', [23]),
                                    ('textFontFamily', [23])]:
            with self.assertRaises(TypeError):
                normalizeEngravingDefaultsAttr(wrongAttr, wrongVal)
        with self.assertRaises(ValueError):
            normalizeEngravingDefaultsAttr('textFontFamily', [''])

    def test_normalizeRequest(self):
        with self.assertRaises(TypeError):
            normalizeEngravingDefaults('path')

    def test_normalizeRequestPath(self):
        with self.assertRaises(TypeError):
            normalizeEngravingDefaults(['path'])


if __name__ == '__main__':
    unittest.main()



import unittest
from smufolib import EngravingDefaults

# pylint: disable=C0115, C0116, C0103


class EngravingDefaults(unittest.TestCase):
    def test_clear(self):
        pass

    def test_items(self):
        pass

    def test_keys(self):
        pass

    def test_update(self):
        pass

    def test_values(self):
        pass

    def test_smufl(self):
        pass

    def test_font(self):
        pass

    def test_glyph(self):
        pass

    def test_layer(self):
        pass

    def test_arrowShaftThickness(self):
        pass

    def test_barlineSeparation(self):
        pass

    def test_beamSpacing(self):
        pass

    def test_beamThickness(self):
        pass

    def test_bracketThickness(self):
        pass

    def test_dashedBarlineDashLength(self):
        pass

    def test_dashedBarlineGapLength(self):
        pass

    def test_dashedBarlineThickness(self):
        pass

    def test_hairpinThickness(self):
        pass

    def test_hBarThickness(self):
        pass

    def test_legerLineExtension(self):
        pass

    def test_legerLineThickness(self):
        pass

    def test_lyricLineThickness(self):
        pass

    def test_octaveLineThickness(self):
        pass

    def test_pedalLineThickness(self):
        pass

    def test_repeatBarlineDotSeparation(self):
        pass

    def test_repeatEndingLineThickness(self):
        pass

    def test_slurEndpointThickness(self):
        pass

    def test_slurMidpointThickness(self):
        pass

    def test_staffLineThickness(self):
        pass

    def test_stemThickness(self):
        pass

    def test_subBracketThickness(self):
        pass

    def test_textFontFamily(self):
        pass

    def test_textEnclosureThickness(self):
        pass

    def test_thickBarlineThickness(self):
        pass

    def test_thinBarlineThickness(self):
        pass

    def test_thinThickBarlineSeparation(self):
        pass

    def test_tieEndpointThickness(self):
        pass

    def test_tieMidpointThickness(self):
        pass

    def test_tieMidpointThickness(self):
        pass

    def test_tupletBracketThickness(self):
        pass

    def test_round(self):
        pass

    def test_spaces(self):
        pass


if __name__ == '__main__':
    unittest.main()



# pylint: disable=C0103, C0114
from types import UnionType
from typing import Any, get_args
import difflib

#: Dictionary of error message templates.
ERROR_TEMPLATES = {
    'alphanumericValue': (
        "The value for '{objectName}' must be alphanumeric."
    ),
    'argumentConflict': (
        "The option '{key}' is already added as positional argument or flag."
    ),
    'attributeError': (
        "'{objectName}' has no attribute '{attribute}'."
    ),
    'dependenTypeError': (
        "Expected '{objectName}' to be of type {validTypes} when "
        "{dependency}, but got {value}."
    ),
    'dependenItemsTypeError': (
        "Items in '{objectName}' must be {validTypes} when {dependency}, "
        "not {value}."
    ),
    'duplicateFlags': (
        "Arguments '{argument1}' and '{argument2}' have duplicate short flag: "
        "{flag}."
    ),
    'duplicateItems': (
        "Items in '{objectName}' cannot be duplicates."
    ),
    'emptyValue': (
        "The value for '{objectName}' cannot be empty."
    ),
    'emptyValueItems': (
        "Value items for '{objectName}' cannot be empty."
    ),
    'fileNotFound': (
        "The file or directory for '{objectName}' does not exist."
    ),
    'invalidFormat': (
        "The value for '{objectName}' is not correctly formatted."
    ),
    'invalidInitialCharacter': (
        "The value for '{objectName}' must start with a lowercase letter or "
        "number."
    ),
    'itemsTypeError': (
        "Items in '{objectName}' must be {validTypes}, not {value}."
    ),
    'itemsValueError': (
        "Invalid value for item in '{objectName}': {value}."
    ),
    'missingExtension': (
        "The value for '{objectName}' must have a {extension} extension."
    ),
    'missingValue': (
        "Required value(s) for '{objectName}' are missing."
    ),
    'nonIncreasingRange': (
        "The values in '{objectName}' must form an increasing range."
    ),
    'notImplementedError': (
        "The '{objectName}' subclass does not implement this method."
    ),
    'numericValue': (
        "The value for '{objectName}' must be numeric."
    ),
    'recommendScript': (
        " Concider running the script '{scriptName}' before the current "
        "process."
    ),
    'serializationError': (
        "Error serializing JSON data or writing to the file."
    ),
    'singleItem': (
        "'{objectName}' must contain a value pair."
    ),
    'suggestion': (
        " Did you mean '{suggestion}'?"
    ),
    'typeError': (
        "Expected '{objectName}' to be of type {validTypes}, but got {value}."
    ),
    'unicodeOutOfRange': (
        "The value for '{objectName}' is outside the Unicode range "
        "(U+0000 – U+10FFFF)."
    ),
    'urlError': (
        "Could not connect to URL: {url}."
    ),
    'valueError': (
        "Invalid value for '{objectName}': {value}."
    ),
    'valueTooHigh': (
        "The value for '{objectName}' must be {value} or lower."
    ),
    'valueTooLow': (
        "The value for '{objectName}' must be {value} or higher."
    )
}


class URLWarning(Warning):
    """URL connection failure warning."""


def generateErrorMessage(*templateNames: str, **kwargs) -> str:
    """Generate an error message from a template and keyword arguments.

    :param errorTemplate: The error message template string, which
        should contain placeholders for keyword arguments.
    :param kwargs: Arbitrary keyword arguments corresponding to the
        placeholders in the template string.
    :raises KeyError: If a placeholder in the template does not have a
        corresponding keyword argument.

    """
    messages = [ERROR_TEMPLATES[n].format(**kwargs) for n in templateNames]
    return ''.join(messages)


def generateTypeError(value: Any,
                      validTypes: type | tuple[type, ...] | UnionType,
                      objectName: str,
                      dependency: str | None = None,
                      items: bool = False) -> str:
    """Generate a :class:`TypeError` message.

    This function generates an error message based on the number of
    valid types. By default, the message is generated from
    :obj:`ERROR_TEMPLATES`:``'typeError'``.

    Examples::

        >>> generateTypeError(123, (str,), 'path')
        Expected 'path' to be of type str, but got int.

    ::

        >>> generateTypeError(123, (str, Path), 'path')
        Expected 'path' to be of type str or Path, but got int.

    ::

        >>> generateTypeError(123, (str, Path, Request), 'path')
        Expected 'path' to be of type str, Path or Request, but got int.

    If `dependency` in not :obj:`None` (or empty) and `items`
    is :obj:`False`, :obj:`ERROR_TEMPLATES`:``'dependentTypeError'``
    is used. If `items` is :obj:`True` and `dependency` is :obj:`None`,
    :obj:`ERROR_TEMPLATES`:``'itemsTypeError'`` is used. If both
    `dependency` is not :obj:`None` and `items` is :obj:`True`,
    :obj:`ERROR_TEMPLATES`:``'dependentItemsTypeError'`` is used.

    :param value: The value to be validated.
    :param validTypes: A :class:`tuple` or :class:`list` of valid types.
    :param objectName: The name of the object being validated.
    :param dependency: Additional substring for dependent type errors.
        template. Defaults to :obj:`None`
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If `items` is :obj:`True` and `value` is not
        an iterable.

    """

    typeNames = _listTypes(validTypes)
    valueType = type(value).__name__

    if items:
        template = 'itemsTypeError'
    elif dependency:
        template = 'dependenTypeError'
    elif items and dependency:
        template = 'dependenItemsTypeError'
    else:
        template = 'typeError'

    return generateErrorMessage(
        template,
        validTypes=typeNames,
        objectName=objectName,
        dependency=dependency,
        value=valueType
    )


def validateType(value: Any,
                 validTypes: type | tuple[type, ...] | UnionType,
                 objectName: str,
                 items=False) -> None:
    """Validate if a value matches any of the valid types.

    :param value: The value to be validated.
    :param validTypes: A single valid type or :class:`tuple` of valid
        types.
    :param objectName: The name of the object being validated.
    :param items: Whether to validate `value` items rather than `value`
        itself. Defaults to :obj:`False`.
    :raises TypeError: If `value` does not match any of the valid
        types.
    :raises ValueError: If  `items` is :obj:`True` and any `value` item
        does not match any of the valid types.

    """
    if isinstance(validTypes, type):
        validTypes = (validTypes,)

    elif isinstance(validTypes, UnionType):
        validTypes = get_args(validTypes)

    if not any(isinstance(value, t) for t in validTypes):
        if items:
            raise ValueError(
                generateTypeError(value, validTypes, objectName, items=items)
            )
        raise TypeError(generateTypeError(value, validTypes, objectName))


def suggestValue(value: str,
                 possibilities: list[str] | tuple[str, ...],
                 objectName: str,
                 cutoff: float=0.6,
                 items=False) -> str:
    """Validate value and uggests a valid close match.

    If `items` is :obj:`True`, an alternate error message template is
    used to validate the values of items within an :term:`iterable`
    value rather than the value itself.

    :param value: The value to be validated.
    :param possibilities: A list of valid possibilities for the value.
    :param objectName: The name of the object being validated.
    :param cutoff: A float value representing the cutoff threshold for
        similarity matches. Defaults to ``0.6``.
    :param items: Whether to use items-specific template. Defaults
        to :obj:`False`.
    :raises ValueError: If the provided value is not found in
        `possibilities` or if `items` is :obj:`True` and `value` is
        not an iterable.

    """
    if value in possibilities:
        return value
    closeMatches = difflib.get_close_matches(value, possibilities, 1, cutoff)
    if closeMatches:
        suggestion = closeMatches[0]
        template = 'itemsValueError' if items else 'valueError'
        raise ValueError(
            generateErrorMessage(
                template, 'suggestion',
                objectName=objectName,
                value=value,
                suggestion=suggestion
            )
        )

    raise ValueError(
        generateErrorMessage(
            'valueError', objectName=objectName, value=value
        )
    )


def _listTypes(types: type | tuple[type, ...] | UnionType) -> str:
    # Represent a series of object type names as a string.

    if isinstance(types, type):
        types = (types,)

    elif isinstance(types, UnionType):
        types = get_args(types)

    if len(types) > 1:
        typeNames = (
            ', '.join(t.__name__ for t in types[:-1])
            + f" or {types[-1].__name__}"
        )
    else:
        typeNames = types[0].__name__
    return typeNames



import unittest
from smufolib.error import (
    VALUE_ERROR,
    VALUE_SUGGESTION,
    MISSING_JSON_EXTENSION,
    TYPE_ERROR,
    URL_ERROR,
    MISSING_URL,
    MISSING_PATH,
    MISSING_PATH_AND_FALLBACK,
    SERIALIZATION_ERROR,
    URLWarning,
    generateErrorMessage,
    validateType,
    suggestValue
)

# pylint: disable=C0115, C0116, C0103


class Error(unittest.TestCase):

    def test_generateErrorMessage(self):
        parameter = 'testParameter'
        value = 'testValue'
        self.assertEqual(generateErrorMessage(
            VALUE_ERROR, parameter=parameter, value=value),
            f"Invalid value for parameter '{parameter}': '{value}'."
        )

    def test_validateType(self):
        pass

    def suggestValue(self):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
# from smufolib import Smufl

# pylint: disable=C0115, C0116, C0103


class Smufl(unittest.TestCase):

    def test_font(self):
        pass

    def test_glyph(self):
        pass

    def test_layer(self):
        pass

    def test_designSize(self):
        pass

    def test_engravingDefaults(self):
        pass

    def test_sizeRange(self):
        pass

    def test_alternates(self):
        pass

    def test_anchors(self):
        pass

    def test_bBox(self):
        pass

    def test_codepoint(self):
        pass

    def test_componentGlyphs(self):
        pass

    def test_range(self):
        pass

    def test_advanceWidth(self):
        pass

    def test_version(self):
        pass

    def test_classes(self):
        pass

    def test_description(self):
        pass

    def test_name(self):
        pass

    def test_isLigature(self):
        pass

    def test_isMember(self):
        pass

    def test_isOptional(self):
        pass

    def test_isRecommended(self):
        pass

    def test_isSalt(self):
        pass

    def test_isSet(self):
        pass

    def test_round(self):
        pass

    def test_toSpaces(self):
        pass

    def test_toUnits(self):
        pass

    def test_spaces(self):
        pass

    def test_base(self):
        pass

    def test_findGlyph(self):
        pass

    def test_suffix(self):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
import random
from pathlib import Path
from smufolib import Font, Glyph, Range, converters
from bin.importID import importID

# pylint: disable=C0115, C0116, C0103


class TestRange(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Populate font with glyphs of different ranges and save to cwd.
        # cls.font = Font()
        # population = range(0xE000, 0xF3FF)
        # for number in random.choices(population, k=20):
        #     glyphName = converters.toUniName(number)
        #     cls.font[glyphName] = Glyph()
        #     cls.font[glyphName].unicode = number
        # cls.font.save('./testRangeFont.ufo')

        # Add SMuFl names to glyphs.
        importID(cls.font, 'name')

    def test_smufl(self):
        pass

    def test_glyph(self):
        pass

    def test_font(self):
        pass

    def test_layer(self):
        pass

    def test_name(self):
        pass

    def test_description(self):
        pass

    def test_glyphs(self):
        pass

    def test_start(self):
        pass

    def test_end(self):
        pass


if __name__ == '__main__':
    unittest.main()
